@@ -1,3 +1,16 @@
+##----------------------------------------------------------------------------
+## Changes file management - ~/lib/Changes.pm
+## Version v0.2.0
+## Copyright(c) 2022 DEGUEST Pte. Ltd.
+## Author: Jacques Deguest <jack@deguest.jp>
+## Created 2022/12/09
+## Modified 2022/12/09
+## All rights reserved
+## 
+## 
+## This program is free software; you can redistribute  it  and/or  modify  it
+## under the same terms as Perl itself.
+##----------------------------------------------------------------------------
 package Changes;
 BEGIN
 {
@@ -36,7 +49,7 @@ BEGIN
     (?<r_tz_space>[[:blank:]\h]+)
     (?<r_tz>\S+)
     /x;
-    our $VERSION = 'v0.1.0';
+    our $VERSION = 'v0.2.0';
 };
 
 use strict;
@@ -45,6 +58,7 @@ use warnings;
 sub init
 {
     my $self = shift( @_ );
+    $self->{defaults}   = undef;
     $self->{elements}   = [];
     $self->{epilogue}   = undef;
     $self->{file}       = undef;
@@ -53,6 +67,7 @@ sub init
     $self->{nl}         = "\n";
     $self->{preamble}   = undef;
     $self->{releases}   = [];
+    $self->{time_zone}  = undef;
     $self->{type}       = undef;
     $self->{wrapper}    = undef;
     $self->{_init_strict_use_sub} = 1;
@@ -60,22 +75,92 @@ sub init
     return( $self );
 }
 
+sub add_epilogue
+{
+    my( $self, $text ) = @_;
+    if( !defined( $text ) || !length( "$text" ) )
+    {
+        return( $self->error( "No text was provided to add an epilogue." ) );
+    }
+    my $elements = $self->elements;
+    my $last = $elements->last;
+    if( defined( $last ) && !$self->_is_a( $last => 'Changes::NewLine' ) )
+    {
+        $elements->push( $self->new_line( nl => ( $self->nl // "\n" ) ) );
+    }
+    $self->epilogue( $text );
+    return( $self );
+}
+
+sub add_preamble
+{
+    my( $self, $text ) = @_;
+    if( !defined( $text ) || !length( "$text" ) )
+    {
+        return( $self->error( "No text was provided to add a premable." ) );
+    }
+    $self->preamble( $text );
+    unless( $text =~ /[\015\012]$/ms )
+    {
+        $self->preamble->append( $self->nl // "\n" );
+    }
+    unless( $text =~ /[\015\012]{2,}$/ms )
+    {
+        $self->preamble->append( $self->nl // "\n" );
+    }
+    return( $self );
+}
+
+sub add_release
+{
+    my $self = shift( @_ );
+    my( $rel, $opts );
+    my $elements = $self->elements;
+    if( scalar( @_ ) == 1 && $self->_is_a( $_[0] => 'Changes::Release' ) )
+    {
+        $rel = shift( @_ );
+        if( $elements->exists( $rel ) )
+        {
+            return( $self->error( "A very same release object with version '", $rel->version, "' is already registered." ) );
+        }
+        my $vers = $rel->version;
+        if( length( "$vers" ) )
+        {
+            my $same = $elements->grep(sub{ $self->_is_a( $_ => 'Changes::Release' ) && $_->version == "$vers" });
+            return( $self->error( "A similar release with version '$vers' is already registered." ) ) if( !$same->is_empty );
+        }
+    }
+    else
+    {
+        $opts = $self->_get_args_as_hash( @_ );
+        if( exists( $opts->{version} ) && defined( $opts->{version} ) && length( "$opts->{version}" ) )
+        {
+            my $vers = $opts->{version};
+            my $same = $elements->grep(sub{ $self->_is_a( $_ => 'Changes::Release' ) && $_->version == "$vers" });
+            return( $self->error( "A similar release with version '$vers' is already registered." ) ) if( !$same->is_empty );
+        }
+        $rel = $self->new_release( %$opts ) || return( $self->pass_error );
+        return( $self->add_release( $rel ) );
+    }
+    $elements->unshift( $self->new_line );
+    $elements->unshift( $rel );
+    return( $rel );
+}
+
 sub as_string
 {
     my $self = shift( @_ );
     my $lines = $self->new_array;
     my $preamble = $self->preamble;
+    my $epilogue = $self->epilogue;
     if( defined( $preamble ) && !$preamble->is_empty )
     {
-        # $lines->push( $preamble->scalar, "\n" );
         $lines->push( $preamble->scalar );
     }
     
     $self->elements->foreach(sub
     {
         $self->message( 4, "Calling as_string on $_" );
-        # XXX
-        $_->debug( $self->debug );
         my $str;
         $str = $_->as_string if( $self->_can( $_ => 'as_string' ) );
         if( defined( $str ) )
@@ -83,10 +168,62 @@ sub as_string
             $lines->push( $str->scalar );
         }
     });
-    # return( $lines->join( "\n" ) . "\n" );
+    if( defined( $epilogue ) && !$epilogue->is_empty )
+    {
+        $lines->push( $epilogue->scalar );
+    }
     return( $lines->join( '' ) );
 }
 
+{
+    *serialize = \&as_string;
+    *serialise = \&as_string;
+}
+
+sub defaults { return( shift->_set_get_hash_as_mix_object( { field => 'defaults', undef_ok => 1 }, @_ ) ); }
+
+sub delete_release
+{
+    my $self = shift( @_ );
+    my $elements = $self->elements;
+    my $removed = $self->new_array;
+    foreach my $rel ( @_ )
+    {
+        if( $self->_is_a( $rel => 'Changes::Release' ) )
+        {
+            my $pos = $elements->pos( $rel );
+            $self->message( 4, "No release object found for object $rel (", overload::StrVal( $rel ), ")" ) if( !defined( $pos ) );
+            my $until = 1;
+            while( defined( $elements->[ $pos + $until ] ) && $self->_is_a( $elements->[ $pos + $until ] => 'Changes::NewLine' ) )
+            {
+                $until++;
+            }
+            $elements->delete( $pos, $until );
+            $removed->push( $rel );
+        }
+        else
+        {
+            my $vers = $rel;
+            if( !defined( $vers ) || !length( "$vers" ) )
+            {
+                warn( "No version provided to remove its corresponding release object.\n" ) if( $self->_warnings_is_enabled );
+                next;
+            }
+            my $found = $elements->grep(sub{ $self->_is_a( $_ => 'Changes::Release' ) && $_->version == $vers });
+            if( $found->is_empty )
+            {
+                next;
+            }
+            $found->foreach(sub
+            {
+                my $deleted = $self->delete_release( $_ );
+                $removed->push( $deleted->list ) if( !$deleted->is_empty );
+            });
+        }
+    }
+    return( $removed );
+}
+
 sub elements { return( shift->_set_get_array_as_object( 'elements', @_ ) ); }
 
 sub epilogue { return( shift->_set_get_scalar_as_object( 'epilogue', @_ ) ); }
@@ -108,6 +245,8 @@ sub freeze
     return( $self );
 }
 
+sub history { return( shift->releases( @_ ) ); }
+
 sub load
 {
     my $this = shift( @_ );
@@ -149,7 +288,16 @@ sub max_width { return( shift->_set_get_number( 'max_width', @_ ) ); }
 sub new_change
 {
     my $self = shift( @_ );
-    my $c = Changes::Change->new( @_ ) ||
+    my $opts = $self->_get_args_as_hash( @_ );
+    my $defaults = $self->defaults;
+    if( defined( $defaults ) )
+    {
+        foreach my $opt ( qw( spacer1 marker spacer2 ) )
+        {
+            $opts->{ $opt } //= $defaults->{ $opt } if( defined( $defaults->{ $opt } ) );
+        }
+    }
+    my $c = Changes::Change->new( $opts ) ||
         return( $self->pass_error( Changes::Change->error ) );
     return( $c );
 }
@@ -157,7 +305,24 @@ sub new_change
 sub new_group
 {
     my $self = shift( @_ );
-    my $g = Changes::Group->new( @_ ) ||
+    my $opts = $self->_get_args_as_hash( @_ );
+    my $defaults = $self->defaults;
+    if( defined( $defaults ) )
+    {
+        my $def = { %$defaults };
+        foreach my $opt ( qw( spacer type ) )
+        {
+            if( !defined( $opts->{ "group_${opt}" } ) && 
+                exists( $def->{ "group_${opt}" } ) && 
+                defined( $def->{ "group_${opt}" } ) && 
+                length( $def->{ "group_${opt}" } ) )
+            {
+                $opts->{ $opt } = CORE::delete( $def->{ "group_${opt}" } );
+            }
+        }
+        $opts->{defaults} //= $def;
+    }
+    my $g = Changes::Group->new( $opts ) ||
         return( $self->pass_error( Changes::Group->error ) );
     return( $g );
 }
@@ -165,6 +330,7 @@ sub new_group
 sub new_line
 {
     my $self = shift( @_ );
+    $self->_load_class( 'Changes::NewLine' ) || return( $self->pass_error );
     my $nl = Changes::NewLine->new( @_ ) ||
         return( $self->pass_error( Changes::NewLine->error ) );
     return( $nl );
@@ -173,11 +339,37 @@ sub new_line
 sub new_release
 {
     my $self = shift( @_ );
-    my $rel = Changes::Release->new( @_ ) ||
+    my $opts = $self->_get_args_as_hash( @_ );
+    my $defaults = $self->defaults;
+    if( defined( $defaults ) )
+    {
+        my $def = { %$defaults };
+        foreach my $opt ( qw( datetime_formatter format spacer time_zone ) )
+        {
+            if( !defined( $opts->{ $opt } ) && 
+                exists( $def->{ $opt } ) && 
+                defined( $def->{ $opt } ) && 
+                length( $def->{ $opt } ) )
+            {
+                $opts->{ $opt } = CORE::delete( $def->{ $opt } );
+            }
+        }
+        $opts->{defaults} //= $def;
+    }
+    my $rel = Changes::Release->new( $opts ) ||
         return( $self->pass_error( Changes::Release->error ) );
     return( $rel );
 }
 
+sub new_version
+{
+    my $self = shift( @_ );
+    $self->_load_class( 'Changes::Version' ) || return( $self->pass_error );
+    my $v = Changes::Version->new( @_ ) ||
+        return( $self->pass_error( Changes::Version->error ) );
+    return( $v );
+}
+
 sub nl { return( shift->_set_get_scalar_as_object( 'nl', @_ ) ); }
 
 sub parse
@@ -197,6 +389,8 @@ sub parse
     # $type is the Changes file type. It contains the value guessed, otherwise it remains undef
     my $type = $self->type;
     my $wrapper = $self->wrapper;
+    my $tz = $self->time_zone;
+    my $defaults = $self->defaults;
     $self->messagef( 4, "Parsing %d lines of data.", $lines->length );
     # Cache it
     unless( defined( $DATETIME_RE ) )
@@ -294,6 +488,8 @@ sub parse
                 container => $self,
                 # Could be undef if this is the last line with no trailing crlf
                 nl => $re->{r_nl},
+                ( defined( $tz ) ? ( time_zone => $tz ) : () ),
+                ( defined( $defaults ) ? ( defaults => $defaults ) : () ),
                 debug => $debug,
             );
             $elements->push( $release );
@@ -343,6 +539,8 @@ sub parse
                 container => $self,
                 # Could be undef if this is the last line with no trailing crlf
                 nl => $re->{r_nl},
+                ( defined( $tz ) ? ( time_zone => $tz ) : () ),
+                ( defined( $defaults ) ? ( defaults => $defaults ) : () ),
                 debug => $debug,
             );
             $elements->push( $release );
@@ -379,6 +577,8 @@ sub parse
                 container => $self,
                 # Could be undef if this is the last line with no trailing crlf
                 nl => $re->{r_nl},
+                ( defined( $tz ) ? ( time_zone => $tz ) : () ),
+                ( defined( $defaults ) ? ( defaults => $defaults ) : () ),
                 debug => $debug,
             );
             $elements->push( $release );
@@ -417,6 +617,7 @@ sub parse
                     type => ( defined( $re->{g_name_colon} ) ? 'colon' : 'bracket' ),
                     # Could be undef if this is the last line with no trailing crlf
                     nl => $re->{g_nl},
+                    ( defined( $defaults ) ? ( defaults => $defaults ) : () ),
                     debug => $debug,
                 );
                 if( !defined( $release ) )
@@ -635,14 +836,60 @@ sub releases
     return( $a );
 }
 
+sub remove_release { return( shift->delete_release( @_ ) ); }
+
+sub time_zone
+{
+    my $self = shift( @_ );
+    if( @_ )
+    {
+        my $v = shift( @_ );
+        if( $self->_is_a( $v => 'DateTime::TimeZone' ) )
+        {
+            $self->{time_zone} = $v;
+        }
+        else
+        {
+            try
+            {
+                $self->_load_class( 'DateTime::TimeZone' ) || return( $self->pass_error );
+                my $tz = DateTime::TimeZone->new( name => "$v" );
+                $self->{time_zone} = $tz;
+            }
+            catch( $e )
+            {
+                return( $self->error( "Error setting time zone for '$v': $e" ) );
+            }
+        }
+        $self->reset(1);
+    }
+    if( !defined( $self->{time_zone} ) )
+    {
+        if( Want::want( 'OBJECT' ) )
+        {
+            require Module::Generic::Null;
+            rreturn( Module::Generic::Null->new( wants => 'OBJECT' ) );
+        }
+        else
+        {
+            return;
+        }
+    }
+    else
+    {
+        return( $self->{time_zone} );
+    }
+}
+
 sub type { return( shift->_set_get_scalar_as_object( 'type', @_ ) ); }
 
+sub wrapper { return( shift->_set_get_code( 'wrapper', @_ ) ); }
+
 sub write
 {
     my $self = shift( @_ );
     my $f = $self->file ||
         return( $self->error( "No Changes file has been set to write to." ) );
-    return( $self->error( "The change file provided cannot be written to." ) ) if( !$f->can_write );
     my $str = $self->as_string;
     return( $self->pass_error ) if( !defined( $str ) );
     if( $str->is_empty )
@@ -657,49 +904,6 @@ sub write
     return( $self );
 }
 
-package
-    Changes::NewLine;
-BEGIN
-{
-    use strict;
-    use warnings;
-    use parent qw( Module::Generic );
-};
-
-sub init
-{
-    my $self = shift( @_ );
-    $self->{nl} = "\n";
-    $self->{raw} = undef;
-    $self->{_init_strict_use_sub} = 1;
-    $self->SUPER::init( @_ ) || return( $self->pass_error );
-    return( $self );
-}
-
-sub as_string
-{
-    my $self = shift( @_ );
-    my $raw = $self->raw;
-    $self->message( 4, "Raw new line was '$raw'" );
-    if( defined( $raw ) && $raw->defined )
-    {
-        return( $raw );
-    }
-    else
-    {
-        my $nl = $self->nl // "\n";
-        return( $self->new_scalar( "$nl" ) );
-    }
-}
-
-sub line { return( shift->_set_get_number( 'line', @_ ) ); }
-
-sub nl { return( shift->_set_get_scalar_as_object( 'nl', @_ ) ); }
-
-sub raw { return( shift->_set_get_scalar_as_object( 'raw', @_ ) ); }
-
-sub wrapper { return( shift->_set_get_code( 'wrapper', @_ ) ); }
-
 1;
 # NOTE: POD
 __END__
@@ -730,12 +934,19 @@ Changes - Changes file management
     $rel->changes->push( $c->new_change(
         text => 'Minor corrections in unit tests',
     ) ) || die( $rel->error );
+    # or
+    my $change = $rel->add_change( text => 'Minor corrections in unit tests' );
+    $rel->delete_change( $change );
+    my $array_object = $c->delete_release( $rel ) ||
+        die( $c->error );
+    say sprintf( "%d releases removed.", $array_object->length );
+    # or $c->remove_release( $rel );
     # Writing to /some/where/else/CHANGES even though we read from /some/where/Changes
     $c->write || die( $c->error );
 
 =head1 VERSION
 
-    v0.1.0
+    v0.2.0
 
 =head1 DESCRIPTION
 
@@ -779,14 +990,86 @@ Provided with some string and an optional hash or hash reference of parameters a
 
 =head1 METHODS
 
+=head2 add_epilogue
+
+Provided with a text and this will set it as the Changes file epilogue, i.e. an optional text that will appear at the end of the Changes file.
+
+If the last element is not a blank line to separate the epilogue from the last release information, then it will be added as necessary.
+
+It returns the current object upon success, or an L<error|Module::Generic/error> upon error.
+
+=head2 add_preamble
+
+Provided with a text and this will set it as the Changes file preamble.
+
+If the text does not have 2 blank new lines at the end, those will be added in order to separate the preamble from the first release line.
+
+It returns the current object upon success, or an L<error|Module::Generic/error> upon error.
+
+=head2 add_release
+
+This takes either an L<Changes::Release> or an hash or hash reference of options required to create one (for that refer to the L<Changes::Release> class), and returns the newly added release object.
+
+The new release object will be added on top of the elements stack with a blank new line separating it from the other releases.
+
+If the same object is found, or an object with the same version number is found, an error is returned, otherwise it returns the release object thus added.
+
 =head2 as_string
 
-Returns a L<string object|Module::Generic::Scalar> representing the entire C<Changes> file. It does so by getting the value set with L<preamble>, and by calling each element stored in L</elements>. Those elements can be L<Changes::Release> and L<Changes::Group> and possibly L<Changes::Change> object. Each one of those element's C<as_string> method will be called to contribute to the overall result.
+Returns a L<string object|Module::Generic::Scalar> representing the entire C<Changes> file. It does so by getting the value set with L<preamble>, and by calling C<as_string> on each element stored in L</elements>. Those elements can be L<Changes::Release> and L<Changes::Group> and possibly L<Changes::Change> object.
 
 If an error occurred, it returns an L<error|Module::Generic/error>
 
 The result of this method is cached so that the second time it is called, the cache is used unless there has been any change.
 
+=head2 defaults
+
+Sets or gets an hash of default values for the L<Changes::Release> or L<Changes::Change> object when it is instantiated upon parsing with L</parse> or by the C<new_release> or C<new_change> method found in L<Changes>, L<Changes::Release> and L<Changes::Group>
+
+Default is C<undef>, which means no default value is set.
+
+    my $ch = Changes->new(
+        file => '/some/where/Changes',
+        defaults => {
+            # for Changes::Release
+            datetime_formatter => sub
+            {
+                my $dt = shift( @_ ) || DateTime->now;
+                require DateTime::Format::Strptime;
+                my $fmt = DateTime::Format::Strptime->new(
+                    pattern => '%FT%T%z',
+                    locale => 'en_GB',
+                );
+                $dt->set_formatter( $fmt );
+                $dt->set_time_zone( 'Asia/Tokyo' );
+                return( $dt );
+            },
+            # No need to provide it if it is just a space though, because it will default to it anyway
+            spacer => ' ',
+            # Not necessary if the custom datetime formatter has already set it
+            time_zone => 'Asia/Tokyo',
+            # for Changes::Change
+            spacer1 => "\t",
+            spacer2 => ' ',
+            marker => '-',
+            max_width => 72,
+            wrapper => $code_reference,
+            # for Changes::Group
+            group_spacer => "\t",
+            group_type => 'bracket', # [Some group]
+        }
+    );
+
+=head2 delete_release
+
+This takes a list of release to remove and returns an L<array object|Module::Generic::Array> of those releases thus removed.
+
+A release provided can either be a L<Changes::Release> object, or a version string.
+
+When removing a release object, it will also take care of removing following blank new lines that typically separate a release from the rest.
+
+If an error occurred, this will return an L<error|Module::Generic/error>
+
 =head2 elements
 
 Sets or gets an L<array object|Module::Generic::Array> of all the elements within the C<Changes> file. Those elements can be L<Changes::Release>, L<Changes::Group>, L<Changes::Change> and C<Changes::NewLine> objects.
@@ -802,6 +1085,12 @@ Sets or gets the text of the epilogue. An epilogue is a chunk of text, possibly
 
 Sets or gets the file path of the Changes file. This returns a L<file object|Module::Generic::File>
 
+=for Pod::Coverage freeze
+
+=head2 history
+
+This is an alias for L</releases> and returns an L<array object|Module::Generic::Array> of L<Changes::Release> objects.
+
 =head2 max_width
 
 Sets or gets the maximum line width for a change inside a release. The line width includes an spaces at the beginning of the line and not just the text of the change itself.
@@ -841,6 +1130,12 @@ Returns a new L<Changes::Release> object, passing it any parameters provided.
 
 If an error occurred, it returns an L<error object|Module::Generic/error>
 
+=head2 new_version
+
+Returns a new C<Changes::Version> object, passing it any parameters provided.
+
+If an error occurred, it returns an L<error object|Module::Generic/error>
+
 =head2 nl
 
 Sets or gets the new line character, which defaults to C<\n>
@@ -861,13 +1156,43 @@ Sets or gets the text of the preamble. A preamble is a chunk of text, possibly m
 
 Read only. This returns an L<array object|Module::Generic::Array> containing all the L<release objects|Changes::Release> within the Changes file.
 
+=head2 remove_release
+
+This is an alias for L</delete_release>
+
+=head2 serialise
+
+This is an alias for L</as_string>
+
+=head2 serialize
+
+This is an alias for L</as_string>
+
+=head2 time_zone
+
+Sets or gets a time zone to use for the release date. A valid time zone can either be an olson time zone string such as C<Asia/Tokyo>, or an L<DateTime::TimeZone> object.
+
+If set, it will be passed to all new L<Changes::Release> object upon parsing with L</parse>
+
+It returns a L<DateTime::TimeZone> object upon success, or an L<error|Module::Generic/error> if an error occurred.
+
 =head2 type
 
 Sets or get the type of C<Changes> file format this is.
 
 =head2 wrapper
 
-Sets or gets a code reference as a callback mechanism to return a properly wrapped change text. This allows flexibility beyond the default use of L<Text::Wrap> and L<Text::Format> by L<Changes::Change>
+Sets or gets a code reference as a callback mechanism to return a properly wrapped change text. This allows flexibility beyond the default use of L<Text::Wrap> and L<Text::Format> by L<Changes::Change>.
+
+If set, this is passed by L</parse> when creating L<Changes::Change> objects.
+
+See L<Changes::Change/as_string> for more information.
+
+=head2 write
+
+This will open the file set with L</file> in write clobbering mode and print out the result from L</as_string>.
+
+It returns the current object upon success, and an L<error|Module::Generic/error> if an error occurred.
 
 =head1 AUTHOR
 
@@ -875,7 +1200,7 @@ Jacques Deguest E<lt>F<jack@deguest.jp>E<gt>
 
 =head1 SEE ALSO
 
-L<Changes::Release>, L<Changes::Group>, L<Changes::Change>
+L<Changes::Release>, L<Changes::Group>, L<Changes::Change>, L<Changes::Version>, L<Changes::NewLine>
 
 =head1 COPYRIGHT & LICENSE
 
